
// include core Wiring API

// include this library's description file
#include "ArCan1.h"
// include description files for other libraries used (if any)
//#include "HardwareSerial.h"

// Sleep Mode and Power Mode
#include <avr/sleep.h>
#include <avr/power.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <avr/sfr_defs.h>

//Incluimos la libreria SPI
#include <SPI.h>



// Constructor /////////////////////////////////////////////////////////////////
// Function that handles the creation and setup of instances
tArCan::tArCan(void) {}
// Public Methods //////////////////////////////////////////////////////////////
// Functions available in Wiring sketches, this library, and other libraries
//Escribir un Byte por SPI
byte tArCan::spi_putc( byte data ) {
    SPDR = data;
	
    // espera hasta que se ha enviado el dato
    while( !( SPSR & (1<<SPIF) ) )
    delay(1);
    return SPDR;
}

//***************************************************************************
//Escribir un registro del mcp2515
void tArCan::write_register( byte direction,byte data) {
	digitalWrite(ArCan_CS,LOW);
	
    spi_putc(SPI_WRITE);
    spi_putc(direction);
    spi_putc(data);

    digitalWrite(ArCan_CS,HIGH);
}

void tArCan::write_registers( byte direction, byte values[], const byte length) {
    byte i;

    digitalWrite(ArCan_CS,LOW);

    spi_putc(SPI_WRITE);
    spi_putc(direction);

    for (i=0; i<length; i++) {
        spi_putc(values[i]);
    }

    digitalWrite(ArCan_CS,HIGH);
}


//Leer un registros del mcp2515
byte tArCan::read_register(byte direction) {
    byte data;
    digitalWrite(ArCan_CS,LOW);

    spi_putc(SPI_READ);
    spi_putc(direction);

    data = spi_putc(0xff);   //Envia un dummy

    digitalWrite(ArCan_CS,HIGH);
    return data;
}

//Leer unos registros del mcp2515
void tArCan::read_registers(byte direction, byte values[], const byte length) {
    byte i;

    digitalWrite(ArCan_CS,LOW);

    spi_putc(SPI_READ);
    spi_putc(direction);

    for (i=0; i<length; i++) {
        values[i] = spi_putc(0xff);
    }

    digitalWrite(ArCan_CS,HIGH);
}

//***************************************************************************
//Modificar un bit del mcp2515
void tArCan::bit_modify(byte direction, byte mask, byte data) {
    digitalWrite(ArCan_CS,LOW);

    spi_putc(SPI_BIT_MODIFY);
    spi_putc(direction);
    spi_putc(mask);
    spi_putc(data);

    digitalWrite(ArCan_CS,HIGH);
}
//***************************************************************************
//Consultar el estado del mcp2515 (2 tipos de consulta SPI_RX_STATUS o SPI_READ_STATUS
byte tArCan::read_status(byte type) {
    byte data;

    digitalWrite(ArCan_CS,LOW);

    spi_putc(type);
    data = spi_putc(0xff);  //Envia un dummy
	spi_putc(0xff);  //Envia un dummy

    digitalWrite(ArCan_CS,HIGH);

    return data;
}
//***************************************************************************
//Inicializa el mcp2515
boolean tArCan::init(uint32_t filter, uint32_t mask_ext, uint32_t mask_std) {
	
	//Inicializamos el puerto SPI
	SPI.begin();
	
    write_register(CANINTE,0);
	
    // Hacemos un reset por soft al mcp2515
    // Despues del reset entra en modo de configuracion
    digitalWrite(ArCan_CS,LOW);
    spi_putc(SPI_RESET);
    digitalWrite(ArCan_CS,HIGH);
	
    // Esperamos un poco para que se estabilize el mcp2515 despues del reset
    delay(20);

    // load CNF1..3 Register
    digitalWrite(ArCan_CS,LOW);
    spi_putc(SPI_WRITE);
    spi_putc(CNF3);

    spi_putc((1<<PHSEG21));
    spi_putc((1<<BTLMODE)|(1<<PHSEG11));
    spi_putc((1<<BRP2)|(1<<BRP1)|(1<<BRP0));

    // Activamos la interrupcion
    spi_putc((1<<RX1IE)|(1<<RX0IE));
    digitalWrite(ArCan_CS,HIGH);

    // Testeamos que la escritura se ha producido correctamente para verificar que el mcp2515 es accesible
    if (read_register(CNF1) != ((1<<BRP2)|(1<<BRP1)|(1<<BRP0))) {
        return false;
    }

    // Para que D2 y D3 se apagen cuando los buffers se llenen
    write_register(BFPCTRL, (1<<B0BFM)|(1<<B1BFM)|(1<<B0BFE)|(1<<B1BFE));

    // Configuramos TXnRTS como entradas digitales
    write_register(TXRTSCTRL, 0);

    
	

	
    // Activamos los filtros en la recepcion de mensajes válidos tanto estandar como extendidos
    write_register(RXB0CTRL, (1<<BUKT));
    write_register(RXB1CTRL, (0<<RXM1) | (0<<RXM0));

	Serial.print("RXB0CTRL  ");Serial.println(read_register(RXB0CTRL),BIN);
	Serial.print("RXB1CTRL  ");Serial.println(read_register(RXB1CTRL),BIN);
	Serial.print("CNF1  ");Serial.println(read_register(CNF1),BIN);
	Serial.print("CNF2  ");Serial.println(read_register(CNF2),BIN);
	Serial.print("CNF3  ");Serial.println(read_register(CNF3),BIN);
	
	
	
	//Establecemos el filtro y la mascara
    setFilter(filter);
    setMask(mask_ext, mask_std);
	
    // Cambiamos al modo Normal
    write_register(CANCTRL, 0);
    write_register(CANINTE, 1);
	
    return true;
}
//***************************************************************************
// Chequea si existe algun mensaje nuevo en espera devuelve 1 si existe o 0 si no hay
boolean tArCan::check_message(void) {
	Serial.println(digitalRead(ArCan_INT));
    return (digitalRead(ArCan_INT));
}

//***************************************************************************
//Chequea si el buffer esta vacio
boolean tArCan::check_free_buffer(void) {
    byte status = read_status(SPI_READ_STATUS);

    if ((status & 0x54) == 0x54) {
        // verifica todos los buffers
        return false;
    }

    return true;
}

//***************************************************************************
//Cogemos el mensaje CAN
byte tArCan::get_message(tCAN *msje) {
	//El dispositivo se despierta en modo listen-only, tiene que ponerse en modo normal
	if((read_register(CANSTAT)>>5)==3){
		mode_normal();
	}
	
	// Se lee el estado para comprabar en que buffer se encuentra el mensaje
    byte status = read_status(SPI_RX_STATUS);
    byte addr;
	
    if (bit_is_set(status,6)) {
        // Si el mensaje esta en el Buffer 1
        addr = SPI_READ_RX;
    }
    else if (bit_is_set(status,7)) {
        // Si el mensaje esta en el Buffer 2
        addr = SPI_READ_RX | 0x04;
    }
    else {
        // Error, el mensaje no esta disponible
        return 0;
    }
	
	
    digitalWrite(ArCan_CS,LOW);
    spi_putc(addr);
	

	
	// leemos la id y los dos bits bajos de la eid, si tiene bit y srr(remota)
	msje->arb_field.id = static_cast<uint32_t>(spi_putc(0xFF))<<24; //ID10..ID3
	msje->arb_field.id |= static_cast<uint32_t>(spi_putc(0xFF))<<16; //ID2..ID0, EID17, EID16(en caso de uso) 
	
	//leemos si es una trama extendida
	msje->arb_field.ide = (bit_is_set(status, 4)) ? 1 : 0;

	
	if(msje->arb_field.ide){
			//leemos EID
			msje->arb_field.id |= static_cast<uint32_t>(spi_putc(0xFF))<<8;//EID15..EID8
			msje->arb_field.id |= spi_putc(0xFF);						   //EID7..EID0		
	}else{
			spi_putc(0xFF);							//no vamos a usar EID
			spi_putc(0xFF);
	}

    // Leemos el DLC
    byte length = spi_putc(0xff) & 0x0f;   //solo nos quedamos con la longitud


    msje->length = length;
    msje->arb_field.rtr = (bit_is_set(status, 3)) ? 1 : 0;

    for (byte i=0;i<length;i++) {
        msje->data[i] = spi_putc(0xff);	
    }
	
    digitalWrite(ArCan_CS,HIGH);
	

    // borramos los flag de las interrupciones
    if (bit_is_set(status, 6)) {
        bit_modify(CANINTF, (1<<RX0IF), 0);
    }
    else {
        bit_modify(CANINTF, (1<<RX1IF), 0);
    }
	
    return status;
}
//***************************************************************************
//Enviamos el mensaje CAN
byte tArCan::send_message(tCAN *msje) {
    byte status = read_status(SPI_READ_STATUS);

    /* Byte de Estado:
     *
     * Bit  Funcion
     *  2   TXB0CNTRL.TXREQ
     *  4   TXB1CNTRL.TXREQ
     *  6   TXB2CNTRL.TXREQ
     *
     *Con esto conseguimos saber que buffer esta libre para enviar el mensaje
     */	
	
    byte address;
    if (bit_is_clear(status, 2)) {
        address = 0x00;
    }
    else if (bit_is_clear(status, 4)) {
        address = 0x02;
    }
    else if (bit_is_clear(status, 6)) {
        address = 0x04;
    }
    else {
        // Todos los buffer estan usados no se puede enviar mensajes
        //devuelve un 0xFF si no se puede enviar
        return 0;
    }

    digitalWrite(ArCan_CS,LOW);
    spi_putc(SPI_WRITE_TX | address);
	
	spi_putc((msje->arb_field.id >> 24) & 0xFF);				//Primero se envia la parte alta ID10....ID3
	spi_putc((msje->arb_field.id >> 16) & 0xFF);				//Despues enviamos la parte baja ID2....ID0, con el bit EXIDE(en caso de uso), con EID17  EID16(en caso de uso)
	//Si es trama extendida
	if(msje->arb_field.ide){		
		spi_putc((msje->arb_field.id >> 8) & 0xFF);				//Enviamos los bits EID15....EID8
		spi_putc(msje->arb_field.id & 0xFF);					//Enviamos los bits EID7....EID0	
	//Si es trama estandar	
	}else{
		spi_putc(0);    								   		//Como no vamos a usar la ID Extendida la ponemos a 0 (parte alta)
		spi_putc(0);            						   		//(parte baja)		
	}
	
	byte length =msje->length & 0x0F;
	
	if(msje->arb_field.rtr){
		spi_putc((1<<RTR) | length);
	}else{
		// se envia la longitud del mensaje
        spi_putc(length);

        // ahora enviamos los datos
        for (byte i=0;i<length;i++) {
            spi_putc(msje->data[i]);
        }
	}

	
    digitalWrite(ArCan_CS,HIGH);

    delay(10);

    // send message
    digitalWrite(ArCan_CS,LOW);
    address = (address == 0) ? 1 : address;
    spi_putc(SPI_RTS | address);
    digitalWrite(ArCan_CS,HIGH);
	
    return address;
}

//***************************************************************************
//Imprime por el puerto serie el mensaje CAN
void tArCan::print_message(tCAN *msje) {

		uint16_t source_id;
		uint16_t destination_id;
		packet_type type;
		uint8_t sequence;
		uint8_t sequence_ack;
		if(msje->arb_field.ide){
			unpack_ext(msje->arb_field.id, source_id, destination_id, type, sequence, sequence_ack);
			Serial.print("id_from: ");
			Serial.println(source_id,DEC);
			Serial.print("id_to: ");
			Serial.println(destination_id,DEC);		
			Serial.print("type: "); Serial.println(type);
			Serial.print("num_seq: ");
			Serial.println(sequence,DEC);
			Serial.print("sequence_ack: ");
			Serial.println(sequence_ack,DEC);
			Serial.println("Extendido");
			
		}else{
			unpack_stand(msje->arb_field.id,source_id,destination_id,sequence);
			Serial.print("id_from: ");
			Serial.println(source_id,DEC);
			Serial.print("id_to: ");
			Serial.println(destination_id,DEC);
			Serial.print("num_seq: ");
			Serial.println(sequence,DEC);
			Serial.println("Standard");
			
		}
		
    Serial.print("Longitud: ");
    Serial.println(msje->length,BIN);
    Serial.print("rtr: ");
    Serial.println(msje->arb_field.rtr,BIN);
	Serial.print("ide: ");
    Serial.println(msje->arb_field.ide,BIN);
	
	
    if (!msje->arb_field.rtr) {
        Serial.print("Datos: ");
        for (byte i = 0; i < msje->length; i++) {
			Serial.print((char)(msje->data[i]));		
        } 

    }
	Serial.println();
	
}

void tArCan::mode_loopback(void) {
    bit_modify(CANCTRL, (1<<REQOP2)|(1<<REQOP1)|(1<<REQOP0), (1<<REQOP1));
	while((read_register(CANSTAT)>>5)!=2){
	}
}

void tArCan::mode_normal(void) {
    bit_modify(CANCTRL, (1<<REQOP2)|(1<<REQOP1)|(1<<REQOP0), 0);	
	while((read_register(CANSTAT)>>5)!=0){
	}
	Serial.println("Mode Normal");
}



void tArCan::setFilter(uint32_t filter){

    digitalWrite(ArCan_CS,LOW);
    spi_putc(SPI_RESET);
    digitalWrite(ArCan_CS,HIGH);

    // Esperamos un poco para que se estabilize el mcp2515 despues del reset
    delay(20);
	
	uint32_t filter_ext=(filter&0x3FF)<<9 |(static_cast<uint32_t>(1)<<19);
	byte BufferFilterExt[4];
    BufferFilterExt[0] = (byte) (filter_ext >> 24) & 0xFF;
    BufferFilterExt[1] = (byte) (filter_ext >> 16) & 0xFF;
    BufferFilterExt[2] = (byte) (filter_ext >> 8) & 0xFF;
    BufferFilterExt[3] = (byte) filter_ext & 0xFF;
	
	uint32_t filter_std=(filter&0xF)<<21;
	//Serial.print();Serial.println();
    byte BufferFilterStd[4];
    BufferFilterStd[0] = (byte) (filter_std >> 24) & 0xFF;
    BufferFilterStd[1] = 0;
    BufferFilterStd[2] = 0;
    BufferFilterStd[3] = 0;
	
	//Mascara 0 Filtros 0 y 1
    write_registers(RXF0SIDH, BufferFilterExt, 4);
    write_registers(RXF1SIDH, BufferFilterExt, 4);
	
	//Mascara 1 Filtros 2,3,4,5
    write_registers(RXF2SIDH, BufferFilterExt, 4);
    write_registers(RXF3SIDH, BufferFilterExt, 4);
    write_registers(RXF4SIDH, BufferFilterExt, 4);
    write_registers(RXF5SIDH, BufferFilterExt, 4);

	byte BufferAux[4];
	read_registers(RXF0SIDH,BufferAux,4);
	Serial.print("RXF0SIDH:"); Serial.println(BufferAux[0],BIN);
	Serial.print("RXF0SIDH:"); Serial.println(BufferAux[1],BIN);
	Serial.print("RXF0SIDH:"); Serial.println(BufferAux[2],BIN);
	Serial.print("RXF0SIDH:"); Serial.println(BufferAux[3],BIN);
	
	read_registers(RXF1SIDH,BufferAux,4);
	Serial.print("RXF1SIDH:"); Serial.println(BufferAux[0],BIN);
	Serial.print("RXF1SIDH:"); Serial.println(BufferAux[1],BIN);
	Serial.print("RXF1SIDH:"); Serial.println(BufferAux[2],BIN);
	Serial.print("RXF1SIDH:"); Serial.println(BufferAux[3],BIN);
	
	read_registers(RXF2SIDH,BufferAux,4);
	Serial.print("RXF2SIDH:"); Serial.println(BufferAux[0],BIN);
	Serial.print("RXF2SIDH:"); Serial.println(BufferAux[1],BIN);
	Serial.print("RXF2SIDH:"); Serial.println(BufferAux[2],BIN);
	Serial.print("RXF2SIDH:"); Serial.println(BufferAux[3],BIN);
	
	read_registers(RXF3SIDH,BufferAux,4);
	Serial.print("RXF3SIDH:"); Serial.println(BufferAux[0],BIN);
	Serial.print("RXF3SIDH:"); Serial.println(BufferAux[1],BIN);
	Serial.print("RXF3SIDH:"); Serial.println(BufferAux[2],BIN);
	Serial.print("RXF3SIDH:"); Serial.println(BufferAux[3],BIN);	
	
	read_registers(RXF4SIDH,BufferAux,4);
	Serial.print("RXF4SIDH:"); Serial.println(BufferAux[0],BIN);
	Serial.print("RXF4SIDH:"); Serial.println(BufferAux[1],BIN);
	Serial.print("RXF4SIDH:"); Serial.println(BufferAux[2],BIN);
	Serial.print("RXF4SIDH:"); Serial.println(BufferAux[3],BIN);	
	
	read_registers(RXF5SIDH,BufferAux,4);
	Serial.print("RXF5SIDH:"); Serial.println(BufferAux[0],BIN);
	Serial.print("RXF5SIDH:"); Serial.println(BufferAux[1],BIN);
	Serial.print("RXF5SIDH:"); Serial.println(BufferAux[2],BIN);
	Serial.print("RXF5SIDH:"); Serial.println(BufferAux[3],BIN);	
	
    // Cambiamos al modo Normal
    write_register(CANCTRL, 0);
}

void tArCan::setMask(uint32_t mask_ext, uint32_t mask_std){
    digitalWrite(ArCan_CS,LOW);
    spi_putc(SPI_RESET);
    digitalWrite(ArCan_CS,HIGH);

    // Esperamos un poco para que se estabilize el mcp2515 despues del reset
    delay(20);

    //Mascara extendida
	byte BufferMaskExt[4];
    BufferMaskExt[0] = (byte) (mask_ext >> 24) & 0xFF;
    BufferMaskExt[1] = (byte) (mask_ext >> 16) & 0xFF;
    BufferMaskExt[2] = (byte) (mask_ext >> 8) & 0xFF;
    BufferMaskExt[3] = (byte) mask_ext & 0xFF;
    //write_registers(RXM0SIDH, BufferMaskStand, 4);
	write_registers(RXM1SIDH, BufferMaskExt, 4);
	
	//Mascara estandard
	byte BufferMaskStand [4];
    BufferMaskStand[0] = (byte) ((mask_std >> 24) & 0xFF);
    BufferMaskStand[1] = 0;
	BufferMaskStand[2] = 0;
    BufferMaskStand[3] = 0;
	//write_registers(RXM1SIDH, BufferMaskStand, 4);
   write_registers(RXM0SIDH, BufferMaskExt, 4);
	
	byte BufferAux[4];
	read_registers(RXM0SIDH,BufferAux,4);
	Serial.print("RXM0SIDH  "); Serial.println(BufferAux[0],BIN);
	Serial.print("RXM0SIDH  "); Serial.println(BufferAux[1],BIN);
	Serial.print("RXM0SIDH  "); Serial.println(BufferAux[2],BIN);
	Serial.print("RXM0SIDH  "); Serial.println(BufferAux[3],BIN);
	
	read_registers(RXM1SIDH,BufferAux,4);
	Serial.print("RXM1SIDH  "); Serial.println(BufferAux[0],BIN);
	Serial.print("RXM1SIDH  "); Serial.println(BufferAux[1],BIN);
	Serial.print("RXM1SIDH  "); Serial.println(BufferAux[2],BIN);
	Serial.print("RXM1SIDH  "); Serial.println(BufferAux[3],BIN);
	

    // Cambiamos al modo Normal
    write_register(CANCTRL, 0);
}


void tArCan::mode_sleep(void) {
	bit_modify(CANCTRL, (1<<REQOP2)|(1<<REQOP1)|(1<<REQOP0), (1<<REQOP0));
	//Habilitamos la interrupcion de wake-up
	write_register(CANINTE, (1<<WAKIE));
	while((read_register(CANSTAT)>>5)!=1){
	}
}
uint32_t tArCan::pack_ext(uint16_t source_id, uint16_t destination_id, packet_type type_of, uint8_t sequence, uint8_t sequence_ack){
	uint32_t buffer=0;

	buffer  = (static_cast<uint32_t>(source_id)<<22) | (static_cast<uint32_t>(1)<<19); //ID10...ID1 con bit EXIDE
	buffer |= (static_cast<uint32_t>(destination_id)<<9);		 					   //ID0, EID17..EID9
	buffer |= (static_cast<uint32_t>(type_of)<<6);			 						   	   //EID8..EID6
	buffer |= (static_cast<uint32_t>(sequence)<<3);			 						   //EID5..EID3
	buffer |= (static_cast<uint32_t>(sequence_ack&0x7));			 			       //EID2..EID0

	return buffer;
}
void tArCan::unpack_ext(uint32_t buffer,uint16_t& source_id, uint16_t& destination_id, packet_type& type_of, uint8_t& sequence, uint8_t& sequence_ack){
	source_id=(buffer>>22)&0x3FF;
	destination_id=(buffer>>9)&0x3FF;
	type_of=get_type((buffer>>6)&0x7);
	sequence=(buffer>>3)&0x7;
	sequence_ack=buffer&0x7;
}
uint32_t tArCan::pack_stand(uint16_t source_id, uint16_t destination_id,uint8_t sequence){
	uint32_t buffer=0;

	buffer = (static_cast<uint32_t>(source_id)<<32-4);		     //ID10...ID7 
	buffer |= (static_cast<uint32_t>(destination_id)<<32-8);	 //ID6...ID3
	buffer |= (static_cast<uint32_t>(sequence)<<32-11);
	
	return buffer;
}

void tArCan::unpack_stand(uint32_t buffer,uint16_t& source_id, uint16_t& destination_id, uint8_t& sequence){
	source_id=(buffer>>32-4)&0xF;
	destination_id=(buffer>>32-8)&0xF;
	sequence=(buffer>>32-11)&0x7;
}

tArCan::packet_type tArCan::get_type(uint8_t type_of){
	switch(type_of){
		case 0: return tArCan::data; break;
		case 1: return tArCan::ack; break;
		case 2: return tArCan::nack; break;
	}
}
