/****************************************************************************
 *
 * Proyecto ArCan
 *
 * Copyright (C) 2008-2009 RaÃºl Milla PÃ©rez raulmp(at)arcan.es
 *          http://www.arcan.es
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 ****************************************************************************/
#ifndef ArCan_h
#define ArCan_h


#include <Arduino.h>
#include "ArCan_def1.h"

// Descripcion del interfaz de la libreria CanBus
class tArCan
{
  // Parte pÃºblica
  public:
	typedef struct{
		  struct{
			  uint32_t id;
			  byte ide : 1;
			  byte rtr : 1;
		  }arb_field;
		  
		  byte length : 4;
		  
		  byte data [8];
	}tCAN;
	enum packet_type {data, ack, nack};
	tArCan(void);
	boolean init(uint32_t filter, uint32_t mask_ext, uint32_t mask_std);
	boolean check_message(void);
	boolean check_free_buffer(void);
	byte get_message(tCAN *msje);
	byte send_message(tCAN *msje);
	void print_message(tCAN *msje);
	
	void mode_loopback(void);
	void mode_normal(void);
	void mode_sleep(void);
    
    
    void setFilter(uint32_t filter);
    void setMask(uint32_t mask_ext, uint32_t mask_std);
	uint32_t pack_ext(uint16_t source_id, uint16_t destination_id, packet_type type_of, uint8_t sequence, uint8_t sequence_ack);
	void unpack_ext(uint32_t buffer, uint16_t& source_id, uint16_t& destination_id, packet_type& type_of, uint8_t& sequence, uint8_t& sequence_ack);
	uint32_t pack_stand(uint16_t source_id, uint16_t destination_id, uint8_t sequence);
	void unpack_stand(uint32_t buffer, uint16_t& source_id, uint16_t& destination_id, uint8_t& sequence);
	packet_type  get_type(uint8_t type_of);
	
  // Parte privada
  private:
    byte spi_putc( byte data );
    void write_register( byte direction, byte data );
    void write_registers( byte direction, byte values[], const byte length);
	
    byte read_register(byte direction);
    void read_registers(byte direction, byte values[], const byte length);

    void bit_modify(byte direction, byte mask, byte data);
    byte read_status(byte type);
};

#endif



