#include "SlidWind.h"
#include <Arduino.h>



#define MAX_RETRIES 5


//Varios niveles de depuración 
//#define DEBUG_1 //Mensajes de depuración de la ventana deslizante
//#define DEBUG_2	//Mensajes de depuración de los timers de envío y recepción
//#define DEBUG_3 //Mensajes de depuración del timeout

tSlidW::tSlidW (void){}

tArCan cans;
int num_buffers;
tTimer timers;
bool enable_sliding_windows; //Habilitar el uso de la ventana

tArCan::tCAN messages;

//Manejo trama CAN
uint16_t source_ids;
uint16_t destination_ids;
tArCan::packet_type type_ofs;
uint8_t sequences;
uint8_t sequence_acks;




bool tSlidW::send_message_wind(tArCan::tCAN message, timeout& tim){
	#ifdef DEBUG_1
		Serial.println(F("////////////ENVIO TRAMA//////////"));
	#endif	
	cans.unpack_ext(message.arb_field.id, source_ids, destination_ids, type_ofs, sequences, sequence_acks);
	if(type_ofs==tArCan::data){//Si la trama es de datos
		if(cans.check_free_buffer()){				//Enviamos la trama
			tim.timer_timeout.start_timer();
			tim.trans_sequence=sequences;
			num_buffers=cans.send_message(&message);
			uint32_t aux = (uint32_t)(tim.timer_timeout.get_time_ms()-tim.timer_timeout.get_value_of_timer());
			tim.trans_typi_dev+=aux*aux;
			tim.trans_tim+=aux;
			tim.trans_num_samples++;
			tim.trans=true;
			#ifdef DEBUG_3
				Serial.println(F("Se envio una trama y se aumento el valor del tiempo de transmision"));
				Serial.print(F("Transmission rate: "));Serial.println((uint32_t) tim.trans_tim);
				Serial.print(F("Numero de muestra de transmision: "));Serial.println(tim.trans_num_samples);
				Serial.print(F("tim.trans_sequence: "));Serial.println(tim.trans_sequence);
				Serial.print(F("Sequences: "));Serial.println(sequences);
			#endif	
			#ifdef DEBUG_1
				cans.print_message(&message);
			#endif	
			if(num_buffers==0){ 
				#ifdef DEBUG_1
					Serial.println(F("Error sending")); 
				#endif	
				return false;
			}
		}else{
			#ifdef DEBUG_1
				Serial.println(F("Error Buffer Full"));
			#endif	
			return false;
		}
	}else{
		return false;
	}
	Serial.print("Se envia trama ");Serial.println(sequences);
	return true;
}
////////////////////////////////////////////////////
/////////////Falta timeout de envio/////////////////
void tSlidW::send_ack_message_wind(recipient& rec){
	#ifdef DEBUG_1
		Serial.println(F("////////////ENVIO ACK//////////"));
	#endif	
	tArCan::tCAN message_auxs;
	for(int i=0;i<NR_BUFS;i++){
		if(rec.input_bufs[i].in_uses){
			message_auxs.arb_field.id=cans.pack_ext(rec.input_bufs[i].destination_id, rec.input_bufs[i].source_id, tArCan::ack, 0, rec.next_confirm);		
			message_auxs.arb_field.ide=1; 									//Envío ack de la secuencia que me llegó con rtr a 1 ya que no vamos a enviar datos
			message_auxs.arb_field.rtr=1;
			message_auxs.length=0;
			if(cans.check_free_buffer()){									//Enviamos la trama
				num_buffers=cans.send_message(&message_auxs);
				Serial.print("Se envía ack ");Serial.println(rec.next_confirm);
				#ifdef DEBUG_1
					cans.print_message(&message_auxs);
					if(num_buffers==0) Serial.println(F("Error sending"));
				#endif
			}else{
				#ifdef DEBUG_1
					Serial.println(F("Error Buffer Full")); 
				#endif	
			}
			break;
		}	
	}	
}

///////////////////////////////////////////////////////////////////
///////////////////////falta timeout de envio//////////////////////
void tSlidW::send_nack_message_wind(entry buffer [NR_BUFS], int seq){
	#ifdef DEBUG_1
		Serial.println(F("////////////ENVIO NACK//////////"));
	#endif	
	tArCan::tCAN message_auxs;
	for(int i=0;i<NR_BUFS;i++){
		if(buffer[i].in_uses){//Cojo el id del destinatario del buffer de salida
			message_auxs.arb_field.id=cans.pack_ext(buffer[i].source_id,buffer[i].destination_id,tArCan::nack,0,seq);//Envío nack de la secuencia con rtr a 1 ya que no vamos a enviar datos
			message_auxs.arb_field.ide=1;
			message_auxs.arb_field.rtr=1;
			message_auxs.length=0;
			if(cans.check_free_buffer()){									//Enviamos la trama
				num_buffers=cans.send_message(&message_auxs);
				#ifdef DEBUG_1
					cans.print_message(&message_auxs);
					if(num_buffers==0) Serial.println(F("Error sending"));
				#endif	
			}else{
				#ifdef DEBUG_1
					Serial.println(F("Error Buffer Full"));
				#endif
			}
		}
	}
}


boolean tSlidW::between(uint8_t lo_lim, uint8_t sequence, uint8_t up_lim){
	return ((lo_lim <= up_lim) && ((sequence<=up_lim)&&(sequence>=lo_lim))) || ((up_lim < lo_lim) && ((sequence<=up_lim)||(sequence>=lo_lim)));
}


void tSlidW::init_sliding_window(uint32_t filter, uint32_t mask_ext, uint32_t mask_std){
	
	//Iniciamos comunicación con el puerto serie
	Serial.begin(250000);
	
	//Arrancamos el temporizador
	timers.init_timer();
	
	if(cans.init(filter,mask_ext,mask_std)){
		Serial.println(F("Starting CANBUS"));
	}else{
		Serial.println(F("CANBUS couldn´t start"));
	}
	
	//Habilitar ventana
	enable_sliding_windows=true;
}	


bool tSlidW::check_input_timers(recipient& rec, sender& sent){
	//Comprobamos que algún timer de recepción haya expirado
	//for(int i=0;i<NR_BUFS;i++){
		/*Serial.print("Timer ");Serial.print(i);Serial.print(" con valor ");Serial.println((uint32_t)rec.timers_in[i].get_value_of_timer());
		Serial.print("Timer 2 ");Serial.println((uint32_t)rec.timers_in[i].get_time_ms());*/
		//if(rec.input_bufs[i].in_uses){		//Si la entrada está en uso tengo que mandar un ack
			#ifdef DEBUG_2
				Serial.print(F("Timer_ack con valor "));Serial.println((uint32_t)rec.timer_ack.get_value_of_timer());
				Serial.print(F("Current time: "));Serial.println((uint32_t)timers.get_time_ms());
			#endif
			if((rec.timer_ack.get_value_of_timer()+(TIMEOUT_ACK_FACTOR*rec.timeout)) <= (timers.get_time_ms())){//Si el timer expiró envío un ack ya que indica que la trama no la he confirmado
				#ifdef DEBUG_2
					Serial.println(F("Timer de entrada expirado."));
				#endif
				//Mando ack y reinicio timer ack
				send_ack_message_wind(rec);
				rec.timer_ack.start_timer();//Inicio el timer de ack para saber cuando fue la ultima vez que envié
				return true;
			}
			//Parte nack
		/*}else{	//Si la entrada no está en uso tengo que mandar un nack si el temporizador expiró
			if(rec.input_bufs[i].no_nack){//Si no se ha mandado nack arrancamos su temporizador
				rec.timers_in[i].start_timer();
			}else{
				if(((rec.timers_in[i].get_value_of_timer()+TIMEOUT_NACK) >= (rec.timers_in[i].get_time_ms()))&&(rec.rec_occupieds>0)){//Si el timer expiró envío un nack ya que indica que la trama no me la han enviado
					Serial.print(F("Timer de entrada "));Serial.print(i);Serial.println(F(" expirado."));
					send_nack_message_wind(sent.out_bufs, i);									//Enviamos trama nack
					rec.timers_in[i].start_timer(); 									//Reiniciamos timer correspondiente recepción ya que hemos enviado el ack
					return true;
				}
			}	
		}*/	
	//}
	return false;
}

bool tSlidW::check_output_timers(recipient& rec, sender& sent){
	bool check=false;
	//Comprobamos que algún timer de envío haya expirado
	for(int i=0;i<NR_BUFS;i++){
		if(sent.out_bufs[i].in_uses){//Si la entrada está en uso compruebo su timer
			#ifdef DEBUG_2
				Serial.print(F("Timer_out "));Serial.print(i);Serial.print(F(" con valor "));Serial.println((uint32_t)sent.timers_out[i].get_value_of_timer());
				Serial.print(F("Current time: "));Serial.println((uint32_t)timers.get_time_ms());
				Serial.print(F("(sent.timers_out[i].get_value_of_timer()+(TIMEOUT_FACTOR*sent.timeout): "));Serial.println((uint32_t)sent.timers_out[i].get_value_of_timer()+(TIMEOUT_FACTOR*sent.timeout));
			#endif	
			if((sent.timers_out[i].get_value_of_timer()+(TIMEOUT_FACTOR*sent.timeout)) <= (timers.get_time_ms())){//Si el timer expiró envío un nack ya que indica que la trama no e la han confirmado
				//if(sent.out_bufs[i].retries<MAX_RETRIES){//Si no se han superado los reintentos maximos de envío
					//sent.out_bufs[i].packet.arb_field.id=cans.pack_ext(sent.out_bufs[i].source_id, sent.out_bufs[i].destination_id, sent.out_bufs[i].type_of,sent.out_bufs[i].sequence, rec.next_confirm);//reconstruyo campo arbitraje del paquete actualizando el sequence_acks
					if(cans.check_free_buffer()){									//Reenviamos la trama
						num_buffers=cans.send_message(&sent.out_bufs[i].packet);
						sent.retries++;
						sent.out_bufs[i].retries++;//Incrementamos los reintentos
						check=true;
						Serial.print("Se reenvía trama ");Serial.println(sent.out_bufs[i].sequence);
						#ifdef DEBUG_1
							Serial.println(F("/////////////////////REENVÍO TRAMA//////////////////////////"));
							cans.print_message(&sent.out_bufs[i].packet);
							if(num_buffers==0) Serial.println(F("Error sending"));
						#endif	
						#ifdef DEBUG_2
							Serial.println(F("Reinicio timer y timer ack"));
						#endif	
						sent.timers_out[i].start_timer();				//Reiniciamos timer respectivo emisor ya que hemos reenviado la trama
						rec.timer_ack.start_timer();
					}else{
						#ifdef DEBUG_1
							Serial.println(F("Error Buffer Full")); 
						#endif	
					}		
					
				//}else{//Si se han superado los reintentos maximos de envío desechamos la trama
				//	sent.out_bufs[i].in_uses=false;
					//sent.out_bufs[i].retries=0;
				//}
			}
		}
	}
	return check;
	
}

bool tSlidW::sent_packet(uint16_t id_source, uint16_t id_destination,uint8_t length, byte data [8], sender& sent, recipient& rec, timeout& tim){
	//Si la ventana está habilitada
	if(enable_sliding_windows){
		////////////////////////////////////////////////////////////
		//////////////////////Parte del emisor//////////////////////
		////////////////////////////////////////////////////////////
		if(((length<=8)&&(length>0))&&(sent.send_occupieds<NR_BUFS)){ //Si hay datos para enviar y tengamos entradas disponibles y la entrada no esté en uso		
			messages=construct_packet(id_source,id_destination, length, data, sent.sequence_now, rec.next_confirm);
			if(!sent.out_bufs[sent.sequence_now&0x3].in_uses){ //Si la entrada no esta ocupada
				#ifdef DEBUG_1
					Serial.println(F("La entrada del emisor no esta siendo ocupada"));
				#endif
				sent.out_bufs[sent.sequence_now&0x3].packet=messages;									//Obtenemos los datos del RS232 y construimos una trama				
				cans.unpack_ext(sent.out_bufs[sent.sequence_now&0x3].packet.arb_field.id,sent.out_bufs[sent.sequence_now&0x3].source_id,sent.out_bufs[sent.sequence_now&0x3].destination_id,sent.out_bufs[sent.sequence_now&0x3].type_of,sent.out_bufs[sent.sequence_now&0x3].sequence,sent.out_bufs[sent.sequence_now&0x3].sequence_ack);
				sent.out_bufs[sent.sequence_now&0x3].in_uses=true;
				sent.out_bufs[sent.sequence_now&0x3].retries=0;				//La entrada está ocupada
				if(sent.send_occupieds==0){ //Si el buffer está vacío
					sent.lo_lim_send_winds=sent.sequence_now;
					sent.up_lim_send_winds=sent.sequence_now;
				}else{
					sent.up_lim_send_winds=(sent.up_lim_send_winds+1)&0x7;
				}
				sent.send_occupieds++;
				if(send_message_wind(sent.out_bufs[sent.sequence_now&0x3].packet, tim)){
					calc_ave_typidev(tim);
					if((tim.transf_ave!=0) && (tim.transf_ave<=sent.timeout)){
						sent.timeout=tim.transf_ave;
						//rec.timeout=4*sent.timeout;
					}
					#ifdef DEBUG_3
						Serial.print(F("Transmission average: "));Serial.println(tim.trans_ave);
						Serial.print(F("sent.timeout: "));Serial.println(sent.timeout);
						Serial.print(F("rec.timeout: "));Serial.println(rec.timeout);
					#endif	
					rec.timer_ack.start_timer();//Inicio el timer de ack para saber cuando fue la ultima vez que envié
					sent.timers_out[sent.sequence_now&0x3].start_timer(); //Inicio el timer respectivo
				}
			}
			sent.sequence_now=(sent.sequence_now+1)&0x7; //Incrementamos la sequencia de cara al siguiente paquete
		}	
		return true;
	}else{//Si la ventana no esta habilitada
		return false;
	}
}

bool tSlidW::receive_packet(recipient& rec, sender& sent, timeout& tim){
	
	bool receive=false;
		
		
	if(enable_sliding_windows){
		////////////////////////////////////////////////////////////
		//////////////////////Parte del receptor////////////////////
		////////////////////////////////////////////////////////////
		timers.start_timer(); //Timer de espera para rececpción
		
		#ifdef DEBUG_2
			Serial.print(F("Timer recepción con valor:"));Serial.println((uint32_t)(timers.get_value_of_timer()+rec.timeout));
			Serial.print(F("Current time: "));Serial.println((uint32_t)timers.get_time_ms());
		#endif
		while((timers.get_value_of_timer()+rec.timeout) >= timers.get_time_ms()){ //Vamos obteniendo tramas timeout msec
			/*Serial.print("timers.get_value_of_timer(): ");Serial.println((uint32_t)timers.get_value_of_timer());
			Serial.print("timers.get_value_of_timer()+rec.timeout: ");Serial.println((uint32_t)timers.get_value_of_timer()+rec.timeout);
			Serial.print("timers.get_time_ms(): ");Serial.println((uint32_t)timers.get_time_ms());*/
				if((cans.get_message(&messages)!=0)){
				#ifdef DEBUG_1
					Serial.println(F("*******************************************************************"));
					Serial.println(F("********************Vamos obteniendo paquetes**********************"));
					Serial.println(F("*******************************************************************"));
				#endif	
				cans.unpack_ext(messages.arb_field.id, source_ids, destination_ids, type_ofs, sequences, sequence_acks);
				#ifdef DEBUG_1
					cans.print_message(&messages);
					Serial.println(F("*******************************************************************"));
					Serial.println(F("********************Se recibio este mensaje************************"));
					Serial.println(F("*******************************************************************"));
				#endif	
				switch(type_ofs){
					case tArCan::nack : //si la trama es de tipo nack reenviamos la trama
						#ifdef DEBUG_1
							Serial.println(F("////////////////ME LLEGA NACK/////////////////////////"));
						#endif
						if(between(sent.out_bufs[sent.lo_lim_send_winds&0x3].sequence, sequence_acks, sent.out_bufs[sent.up_lim_send_winds&0x3].sequence)){
							if(cans.check_free_buffer()){
								//Renviamos la trama correspondiente desde el buffer del emisor
								// las tramas nack no llevan datos con lo que no hay que verificar que están en la ventana deslizante de entrada
								// esto debe tratarse fuera del if(betwee(...)), y debe mirarse en la ventana deslizante de salida
								// mirar que el nro de secuencia se corresponde con una entrada en uso
								num_buffers=cans.send_message(&sent.out_bufs[sequences&0x3].packet);
								cans.print_message(&sent.out_bufs[sequences&0x3].packet);
								sent.timers_out[sequences&0x3].start_timer();	//Reiniciamos el temporizador correspondiente
								#ifdef DEBUG_1
									if(num_buffers==0) Serial.println(F("Error sending"));
								#endif	
							}else{
								#ifdef DEBUG_1
									Serial.println(F("Error Buffer Full"));  
								#endif	
							}
						}	
						break;
					case tArCan::ack : //si la trama es de tipo ack la trama esta confirmada y todos sus números de secuencia anteriores
							#ifdef DEBUG_1
								Serial.println(F("////////////////ME LLEGA ACK/////////////////////////"));
								cans.print_message(&messages);
							#endif
							Serial.print("Me llega confirmación ");Serial.println(sequence_acks);
							check_ack(sent, sequence_acks, tim);
							calc_ave_typidev(tim);
							// if(tim.transf_ave<=rec.timeout){
								// rec.timeout=tim.transf_ave;
							// }
							#ifdef DEBUG_3
								Serial.print("Transference average: ");Serial.println(tim.transf_ave);
								Serial.print("rec.timeout: ");Serial.println(rec.timeout);
							#endif
						break;
					default :	//si la trama es de datos la metemos en el buffer y miramos por si nos confiman otras enviadas
						//Si la sequencia esta entre las esperadas
							if(between(rec.lo_lim_rec_winds,sequences,rec.up_lim_rec_winds)){
								#ifdef DEBUG_1
									Serial.println(F("La sequencia está entre las esperadas"));
								#endif	
								if(rec.input_bufs[sequences&0x3].in_uses){//Si dicha entrada esta en uso quiere decir que se trata de un reenvío y enviamos un ack
									// aquí no es necesario enviar un ack, el ack se envía implícitamente o bien, cuando se envíe un paquete de datos, o 
									// explícitamente cuando se el temporizador de acks se active y haya que enviar un paquete de ack "sin datos"
									// habría que verificar que el paquete recibido coincide con el que tenemos almacenado a efectos de comprobación
									// de errores y depuración.
									if(rec.input_bufs[sequences&0x3].sequence==sequences){
										rec.bads_r++;
										#ifdef DEBUG_1
											Serial.println(F("Me llego la misma trama"));
											cans.print_message(&messages);
											Serial.println(F("//////////////////////////////"));
										#endif		
										check_ack(sent, sequence_acks, tim);
										calc_ave_typidev(tim);
										// if(tim.transf_ave<=rec.timeout){
											// rec.timeout=(int)tim.transf_ave;
										// }
										#ifdef DEBUG_3
											Serial.print("Transference average: ");Serial.println(tim.transf_ave);
											Serial.print("rec.timeout: ");Serial.println(rec.timeout);
										#endif					
									}
								}else{//Si la entrada no está en uso
									#ifdef DEBUG_1
										Serial.println(F("Entrada receptor no en uso"));
									#endif
									rec.goods_r++;
									rec.input_bufs[sequences&0x3].packet=messages;
									#ifdef DEBUG_1
										Serial.println(F("////////////////ME LLEGA PAQUETE/////////////////////////"));
										cans.print_message(&messages);
									#endif	
									cans.unpack_ext(rec.input_bufs[sequences&0x3].packet.arb_field.id, rec.input_bufs[sequences&0x3].source_id, rec.input_bufs[sequences&0x3].destination_id, rec.input_bufs[sequences&0x3].type_of, rec.input_bufs[sequences&0x3].sequence, rec.input_bufs[sequences&0x3].sequence_ack);
									rec.input_bufs[sequences&0x3].in_uses=true;
									rec.input_bufs[sequences&0x3].retries=0;			//Marcamos la entrada como usada
									rec.rec_occupieds++;
									//rec.up_lim_rec_winds=(++rec.up_lim_rec_winds)&0x3;
									Serial.print("Me llega la trama ");Serial.println(sequences);
									if(between(sent.out_bufs[sent.lo_lim_send_winds&0x3].sequence,rec.input_bufs[sequences&0x3].sequence_ack,sent.out_bufs[sent.up_lim_send_winds&0x3].sequence)){//Si nos van a  confirmar tramas y dicha confirmación está en los limites de la ventana de recepción
										check_ack(sent, rec.input_bufs[sequences&0x3].sequence_ack, tim);
										calc_ave_typidev(tim);
										// if(tim.transf_ave<=rec.timeout){
											// rec.timeout=tim.transf_ave;
										// }
										#ifdef DEBUG_3
											Serial.print("Transference average: ");Serial.println(tim.transf_ave);
											Serial.print("rec.timeout: ");Serial.println(rec.timeout);
										#endif
									}
									if(sequences>=((rec.next_confirm+1)&0x7)){ //Si la sequencia actual obtenida es mayor q el valor que yo tengo para confirmar
										//Si la  entrada en el buffer de entrada de mi confirmación esta usada quiere decir que la tengo
										#ifdef DEBUG_1
											Serial.println(F("Incremento next_confirm: "));
											Serial.print(F("Valor next_confirm antes de incrementar: "));Serial.println(rec.next_confirm);
										#endif
										int i,n;
										for(
											i=(rec.next_confirm+1)&0x3, n=0;
											n<NR_BUFS;
											i=(i+1)&0x3, n++
										){ //Recorro el resto de entradas desde mi confirmación para ver las que tengo
											if(!rec.input_bufs[i].in_uses) break;										
											//Si están siendo usadas las confirmo
											rec.next_confirm=(rec.next_confirm+1)&0x7;
											#ifdef DEBUG_1
												Serial.print(F("rec.input_bufs[i].in_uses"));Serial.println(rec.input_bufs[i].in_uses);
												Serial.print(F("Valor de i: "));Serial.println(i);
												Serial.print(F("Valor next_confirm: "));Serial.println(rec.next_confirm);
											#endif	
										}	
									}
								}
							}else{
								rec.bads_r++;
								#ifdef DEBUG_1
									Serial.println(F("La sequencia no está entre las esperadas"));
								#endif							
							}	
						break;
				}
				receive=true;
			}	
		}		
	}else{//Si la ventana no esta habilitada
		return false;
	}
	return receive;
}

tArCan::tCAN tSlidW::construct_packet(uint16_t id_source, uint16_t id_destination, uint8_t length, byte data [8], uint8_t sequence_now, uint8_t next_confirm){
	messages.arb_field.id=cans.pack_ext(id_source, id_destination, tArCan::data, sequence_now, next_confirm);
	messages.arb_field.ide=1;
	messages.arb_field.rtr=0;
	if(length>8){
		messages.length=8;
		for(int i=0;i<8;i++){
			messages.data[i]=data[i];
		}
	}else{
		messages.length=length;
		for(int i=0;i<length;i++){
			messages.data[i]=data[i];
		}
	}		
	return messages;	
	
}

bool tSlidW::check_window_status(int send_occupieds, recipient rec){
	if((rec.rec_occupieds>=NR_BUFS)||(send_occupieds>=NR_BUFS)){	//Buffer de entrada o de salida lleno
		enable_sliding_windows=false;
		rec.next_confirm=0;
		return false;
	}else{
		return true;
	}
}

void tSlidW::print_sender(sender sent){
	Serial.println(F("////////////////////EMISOR///////////////////"));
	Serial.print(F("Limite inferior ventana: "));Serial.println(sent.lo_lim_send_winds);
	Serial.print(F("Limite superior ventana: "));Serial.println(sent.up_lim_send_winds);
	Serial.print(F("Entradas emisor ocupadas: "));Serial.println(sent.send_occupieds);
	Serial.print(F("Sequencia actual: "));Serial.println(sent.sequence_now);
	print_buffer(sent.out_bufs);
}

void tSlidW::print_recipient(recipient rec){
	Serial.println(F("////////////////////RECEPTOR///////////////////"));
	Serial.print(F("Limite inferior ventana: "));Serial.println(rec.lo_lim_rec_winds);
	Serial.print(F("Limite superior ventana: "));Serial.println(rec.up_lim_rec_winds);
	Serial.print(F("Entradas receptor ocupadas: "));Serial.println(rec.rec_occupieds);
	Serial.print(F("Siguiente confirmación: "));Serial.println(rec.next_confirm);
	print_buffer(rec.input_bufs);

}

void tSlidW::print_buffer(entry buffer [NR_BUFS]){
	for(int i=0;i<NR_BUFS;i++){
		if(buffer[i].in_uses){
			Serial.print(F("//////////TRAMA "));Serial.print(i);Serial.println(F(" /////////////"));
			cans.print_message(&buffer[i].packet);
			Serial.print(F("In_use: "));Serial.println(buffer[i].in_uses);
			Serial.print(F("Retries: "));Serial.println(buffer[i].retries);
		}
	}
}

//Comprobar que tengo algo en el lo_lim 
bool tSlidW::get_data(uint16_t& id, uint8_t& length, byte data [8], recipient& rec){
	if(rec.input_bufs[rec.lo_lim_rec_winds&0x3].in_uses){ //Si la entrada está en uso devuelvo los datos
		#ifdef DEBUG_1
			Serial.println(F("///////////////////////////////////////////////////////////////"));
			Serial.println(F("////////////////////////////LIBERO RECEPTOR////////////////////"));
			Serial.println(F("///////////////////////////////////////////////////////////////"));
		#endif	
		//Obtengo el id
		id=rec.input_bufs[rec.lo_lim_rec_winds&0x3].source_id;
		
		//Obtengo la longitud 
		length=rec.input_bufs[rec.lo_lim_rec_winds&0x3].packet.length;
		
		//Obtengo los datos
		copy_data(rec.input_bufs[rec.lo_lim_rec_winds&0x3],data);
		
		//Avanzamos en la ventana liberando
		rec.input_bufs[rec.lo_lim_rec_winds&0x3].in_uses=false;
		rec.lo_lim_rec_winds=(rec.lo_lim_rec_winds+1)&0x7;
		
		rec.up_lim_rec_winds=(rec.up_lim_rec_winds+1)&0x7;
		rec.rec_occupieds--;
		return true;
	}else{ //Si la entrada no está en uso
		#ifdef DEBUG_1
			Serial.println(F("///////////////////////////////////////////////////////////////"));
			Serial.println(F("/////////////////////////NO LIBERO RECEPTOR////////////////////"));
			Serial.println(F("///////////////////////////////////////////////////////////////"));
		#endif	
		return false;
	}	
}

void tSlidW::copy_data(entry buffer, byte data [8]){ 
	for(int i=0;i<buffer.packet.length;i++){
		data[i]=buffer.packet.data[i];
	}
}
void tSlidW::check_ack(sender& sent, uint8_t sequence_acks, timeout& tim){
	#ifdef DEBUG_1
		Serial.println(F("Compruebo confirmación"));
		Serial.print(F("sequence_acks: "));Serial.println(sequence_acks);
		Serial.print(F("LO_LIM_SEND_WINDS: "));Serial.println(sent.lo_lim_send_winds);
		Serial.print(F("UP_LIM_SEND_WINDS: "));Serial.println(sent.up_lim_send_winds);
		Serial.print(F("between(sent.lo_lim_send_winds,sequence_acks,sent.up_lim_send_winds) :"));Serial.println(between(sent.lo_lim_send_winds,sequence_acks,sent.up_lim_send_winds));

	#endif
	if(sent.out_bufs[sequence_acks&0x3].in_uses){
		#ifdef DEBUG_1
			Serial.println(F("Si la entrada emisor está en uso"));
		#endif
		while((sent.lo_lim_send_winds!=sent.up_lim_send_winds)&&(between(sent.lo_lim_send_winds,sequence_acks,sent.up_lim_send_winds))&&(sequence_acks>=0)){ //Mientras los límites sean distintos voy comprobando
			#ifdef DEBUG_1
				Serial.println(F("Entro while"));
				cans.print_message(&sent.out_bufs[sent.lo_lim_send_winds&0x3].packet);
			#endif			
			if(between(sent.lo_lim_send_winds,sequence_acks,sent.up_lim_send_winds)){ //Me estan confirmando mientras la sequencia del limite inferior sea menor a la que me están confirmando
				#ifdef DEBUG_1
					Serial.print(F("LO_LIM_SEND_WINDS: "));Serial.println(sent.lo_lim_send_winds);
					Serial.print(F("UP_LIM_SEND_WINDS: "));Serial.println(sent.up_lim_send_winds);
					Serial.print(F("between(sent.lo_lim_send_winds,sequence_acks,sent.up_lim_send_winds) :"));Serial.println(between(sent.lo_lim_send_winds,sequence_acks,sent.up_lim_send_winds));
					Serial.println(F("Sequencia limite inferior menor que la confirmación y libero"));
				#endif
				#ifdef DEBUG_3
					Serial.print(F("tim.trans_sequence: "));Serial.println(tim.trans_sequence);
					Serial.print(F("sequence_acks:	"));Serial.println(sequence_acks);
					Serial.print(F("Transference_rate:"));Serial.println((uint32_t) tim.transf_tim);					
				#endif
				if(tim.trans_sequence<=sequence_acks){
					uint32_t aux = (uint32_t) (tim.timer_timeout.get_time_ms()-tim.timer_timeout.get_value_of_timer());
					tim.transf_typi_dev += aux*aux;
					tim.transf_tim += aux;
					tim.transf_num_samples++;
					tim.trans=false;
					#ifdef DEBUG_3
						Serial.println(F("Sequencia confirmada, aumento el valor del tiempo de transferencia"));
						Serial.print(F("Transference_rate:"));Serial.println(tim.transf_tim);
					#endif
				}
				Serial.print("Se confirmó la trama ");Serial.println(sent.lo_lim_send_winds);
				sent.out_bufs[sent.lo_lim_send_winds&0x3].in_uses=false;
				/*sent.out_bufs[sent.lo_lim_send_winds].retries=0;
				sent.timers_out[sent.lo_lim_send_winds].reset_timer();*/ //Paramos timer respectivo emisor ya que hemos recibido el ack
				sent.lo_lim_send_winds=(sent.lo_lim_send_winds+1)&0x7;
				sent.send_occupieds--;
				sent.confirms++;
			}
		}
		if((between(sent.lo_lim_send_winds,sequence_acks,sent.up_lim_send_winds))&&(sent.send_occupieds>0)&&(sequence_acks>=0)){ //una vez sean iguales comparo la ultima secuencia si tengo entradas disponibles sino quiere decir que ya las compare
				#ifdef DEBUG_1
					Serial.print(F("LO_LIM_SEND_WINDS: "));Serial.println(sent.lo_lim_send_winds);
					Serial.print(F("UP_LIM_SEND_WINDS: "));Serial.println(sent.up_lim_send_winds);
					Serial.print(F("between(sent.lo_lim_send_winds,sequence_acks,sent.up_lim_send_winds) :"));Serial.println(between(sent.lo_lim_send_winds,sequence_acks,sent.up_lim_send_winds));
					Serial.println(F("Si los limites son iguales y la secuencia es menor que la confirmacion"));
				#endif	
				#ifdef DEBUG_3
					Serial.print(F("tim.trans_sequence: "));Serial.println(tim.trans_sequence);
					Serial.print(F("sequence_acks:	"));Serial.println(sequence_acks);
					Serial.print(F("Transference_rate:"));Serial.println((uint32_t) tim.transf_tim);					
				#endif
				if(tim.trans_sequence<=sequence_acks){
					uint32_t aux = (uint32_t) (tim.timer_timeout.get_time_ms()-tim.timer_timeout.get_value_of_timer());
					tim.transf_typi_dev += aux*aux;
					tim.transf_tim += aux;
					tim.transf_num_samples++;
					tim.trans=false;
					#ifdef DEBUG_3
						Serial.println(F("Sequencia confirmada, aumento el valor del tiempo de transferencia"));
						Serial.print(F("Transference_rate:"));Serial.println(tim.transf_tim);
					#endif
				}
				sent.confirms++;
				Serial.print("Se confirmó la trama ");Serial.println(sent.lo_lim_send_winds);				
				sent.out_bufs[sent.lo_lim_send_winds&0x3].in_uses=false;
				sent.out_bufs[sent.lo_lim_send_winds&0x3].retries=0;
				sent.lo_lim_send_winds=-1;
				sent.up_lim_send_winds=-1;
				sent.send_occupieds=0;
		}
		#ifdef DEBUG_1
			print_sender(sent);
		#endif	
	}	
}

void tSlidW::calc_ave_typidev(timeout& tim){
	if(tim.trans){
		tim.trans_ave=(double)(tim.trans_tim)/(double)(tim.trans_num_samples);
		#ifdef DEBUG_3
			Serial.print(F("Transmission average: "));Serial.println(tim.trans_ave);
			Serial.print(F("tim.trans_typi_dev: "));Serial.println(tim.trans_typi_dev);
			Serial.print(F("tim.trans_num_samples: "));Serial.println(tim.trans_num_samples);
			Serial.print(F("(double)(tim.trans_typi_dev/tim.trans_num_samples): "));Serial.println((double)(tim.trans_typi_dev)/(double)(tim.trans_num_samples));
			Serial.print(F("(tim.trans_ave*tim.trans_ave): "));Serial.println((tim.trans_ave*tim.trans_ave));
		#endif
		tim.trans_dev=(double)(tim.trans_typi_dev)/(double)(tim.trans_num_samples)-(tim.trans_ave*tim.trans_ave);
		tim.trans_dev=sqrt(tim.trans_dev);
		#ifdef DEBUG_3
			Serial.print(F("Transmission typical deviation: "));Serial.println(tim.trans_dev);
		#endif	
	}else{
		tim.transf_ave=(double)(tim.transf_tim)/(double)(tim.transf_num_samples);
		#ifdef DEBUG_3
			Serial.print(F("Transference average: "));Serial.println(tim.transf_ave);
			Serial.print(F("tim.transf_typi_dev: "));Serial.println(tim.transf_typi_dev);
			Serial.print(F("tim.transf_num_samples: "));Serial.println(tim.transf_num_samples);
			Serial.print(F("(double)(tim.transf_typi_dev/tim.transf_num_samples): "));Serial.println((double)(tim.transf_typi_dev)/(double)(tim.transf_num_samples));
			Serial.print(F("(tim.transf_ave*tim.transf_ave): "));Serial.println((tim.transf_ave*tim.transf_ave));
			
		#endif
		tim.transf_dev=(double)(tim.transf_typi_dev)/(double)(tim.transf_num_samples)-(tim.transf_ave*tim.transf_ave);
		tim.transf_dev=sqrt(tim.transf_dev);
		#ifdef DEBUG_3
			Serial.print(F("Transference typical deviation: "));Serial.println(tim.transf_dev);
		#endif
	}
}
bool tSlidW::introduce_data(String& incom){
	if (Serial.available()>0){
		incom=Serial.readStringUntil('\n');
		return true;
	}
	return false;
}
void tSlidW::get_byte_data(String& incom, byte data [8],uint8_t& length){
	if(incom.length()>8){
		for(int i=0;i<8;i++){
			data[i]=incom.charAt(i);
		}
		incom=incom.substring(8,incom.length());
		length=8;
	}else{
		for(int i=0,length=0;i<incom.length();i++,length++){
			data[i]=incom.charAt(i);
		}
		incom="";
	}
}
void tSlidW::print_timeout(timeout tim){
		Serial.print(F("Transmission average: "));Serial.println(tim.trans_ave);
		Serial.print(F("Transmission typical deviation: "));Serial.println(tim.trans_dev);
		Serial.print(F("Transference average: "));Serial.println(tim.transf_ave);
		Serial.print(F("Transference typical deviation: "));Serial.println(tim.transf_dev);
}