#include <Arduino.h>
#include "ArCan1.h"
#include "Timer.h"

#define MAX_SEQ 7  //2^n - 1
#ifdef MAX_SEQ
	#define NR_BUFS ((MAX_SEQ + 1)/2)
#endif

#define TIMEOUT_FACTOR  4//2//3 
#define TIMEOUT_ACK_FACTOR  4//2//3 

class tSlidW{
public: 

	typedef struct{
		//in-use-> saber si la entrada esta siendo usada, retries->numero de reintentos de envio
		bool in_uses=false;
		int retries=0;
		
		tArCan::tCAN packet;
		
		//Manejo trama CAN
		uint16_t source_id;
		uint16_t destination_id;
		tArCan::packet_type type_of;
		uint8_t sequence;
		uint8_t sequence_ack;
		
	}entry;
	
	typedef struct{
		tSlidW::entry 		 out_bufs   [NR_BUFS];
		tTimer				 timers_out	[NR_BUFS];
		int lo_lim_send_winds=-1;		
		int up_lim_send_winds=-1;
		int send_occupieds=0;
		uint8_t sequence_now=0; //Siguiente número de secuencia que tengo que enviar
		double timeout=1000;
		uint32_t retries=0; //Número total de reenvíos realizados
		uint32_t confirms=0; //Número total de paquetes conirmados
	}sender;
	
	typedef struct{
		tSlidW::entry 		 input_bufs	[NR_BUFS];
		tTimer		 		 timer_ack;
		int lo_lim_rec_winds=0;
		int up_lim_rec_winds=3;
		int rec_occupieds=0;
		int next_confirm=-1; //Confirmación que tengo que enviar al emisor
		double timeout=100;
		uint32_t goods_r=0; //Paquetes recibidos con las sequencias esperadas
		uint32_t bads_r=0;	//Paquetes recibidos fuera de las secuencias esperadas
	}recipient;

	
	typedef struct{
		//Manejo TIMEOUT
		tTimer timer_timeout;
		//Transmisión
		uint32_t trans_tim=0;
		uint32_t trans_typi_dev=0;
		double trans_ave=0;
		double trans_dev=0;
		uint32_t trans_num_samples=0;
		uint8_t trans_sequence;
		bool trans;
		//Transferencia
		uint32_t transf_tim=0;
		uint32_t transf_typi_dev=0;
		double transf_ave=0;
		double transf_dev=0;
		uint32_t transf_num_samples=0;
	}timeout;
	
	tSlidW(void);
	
		
	void init_sliding_window(uint32_t filter, uint32_t mask_ext, uint32_t mask_std); 
	bool check_input_timers(recipient& rec, sender& sent);
	bool check_output_timers(recipient& rec, sender& sent);
	bool check_window_status(int send_occupieds, recipient rec);
	bool sent_packet(uint16_t id_source, uint16_t id_destination,uint8_t length, byte data [8], sender& sent, recipient& rec, timeout& tim);
	bool receive_packet(recipient& rec, sender& sent, timeout& tim);
	void print_buffer(entry buffer [NR_BUFS]);
	void print_sender(sender sent);
	void print_recipient(recipient rec);
	bool get_data(uint16_t& id, uint8_t& length, byte data [8], recipient& rec);
	bool introduce_data(String& incom);
	void get_byte_data(String& incom, byte data [8],uint8_t& length);
	void print_timeout(timeout tim);
	
private:	
	void send_ack_message_wind(recipient& rec);
	void send_nack_message_wind(entry buffer [NR_BUFS], int i);
	bool send_message_wind(tArCan::tCAN message, timeout& tim);
	tArCan::tCAN construct_packet(uint16_t id_source, uint16_t id_destination, uint8_t length, byte data [8], uint8_t sequence_now, uint8_t next_confirm);
	boolean between(uint8_t lo_lim, uint8_t up_lim, uint8_t sequence);
	
	void copy_data(entry buffer, byte data [8]);
	void check_ack(sender& sent, uint8_t sequence_acks, timeout& tim);
	
	void calc_ave_typidev(timeout& tim);
	
};